#!/usr/bin/env python
# coding: utf-8

# In[1]:
"""
@Author: Pedro Herrera-Lormendez
"""

import numpy as np
import pandas as pd
import xarray as xr

def eleven_CTs(CT):
    '''
    Computes the reduced 11 circulation types based on their advective
    characteristics

    :params CT: DataArray of the 27 circulation types
    '''
    CT = xr.where( (CT == 11) | (CT==21) | (CT ==1), 1, CT) #NE
    CT = xr.where( (CT == 12) | (CT==22) | (CT==2), 2, CT) #E
    CT = xr.where( (CT == 13) | (CT==23) | (CT==3), 3, CT) #SE
    CT = xr.where( (CT == 14) | (CT==24) | (CT==4), 4, CT) #S
    CT = xr.where( (CT == 15) | (CT==25) | (CT==5), 5, CT) #SW
    CT = xr.where( (CT == 16) | (CT==26) | (CT==6), 6, CT) #W
    CT = xr.where( (CT == 17) | (CT==27) | (CT==7), 7, CT) #NW
    CT = xr.where( (CT == 18) | (CT==28) | (CT==8), 8, CT) #N
    # CT = xr.where(CT == -1, 0, CT)
    # CT = xr.where(CT ==  0, 10, CT)
    CT = xr.where(CT == 20, 9, CT) #C
    return(CT)

# def seasonal_frelative_frequencies(CTs_11, year_init, year_end):
#     '''
#     This function computes the climatological seasonal relative frequencies 
#     of the JK derived 11 CTs

#     :param CTs_11: DataArray of the reduced eleven circulation types
#     :year_init   : str Starting year to computer the climatology
#     :year_end    : str Ending year to compute the climatology
    
#     '''
#     #Reading the daily CTs dataset
#     CT = DS.CT.sel(time = slice(str(year_init)+'-03-01', str(year_end)+'-11-30'))
#     del(DS)
#     LF = xr.where(CT == -1, 1, np.nan)
#     A = xr.where(CT ==  0, 1, np.nan)
#     C = xr.where(CT == 20, 1, np.nan)
#     NE = xr.where( (CT == 11) | (CT==21) | (CT ==1), 1, np.nan)
#     E = xr.where( (CT == 12) | (CT==22) | (CT==2), 1, np.nan)
#     SE = xr.where( (CT == 13) | (CT==23) | (CT==3), 1, np.nan)
#     S = xr.where( (CT == 14) | (CT==24) | (CT==4), 1, np.nan)
#     SW = xr.where( (CT == 15) | (CT==25) | (CT==5), 1, np.nan)
#     W = xr.where( (CT == 16) | (CT==26) | (CT==6), 1, np.nan)
#     NW = xr.where( (CT == 17) | (CT==27) | (CT==7), 1, np.nan)
#     N = xr.where( (CT == 18) | (CT==28) | (CT==8), 1, np.nan)
#     month_length = CT.time.dt.days_in_month
#     len_seasons=month_length.resample(time = '1M').mean().resample(time = 'Q-FEB').sum()
#     # weights = month_length.groupby('time.season') / month_length.groupby('time.season').sum()
#     LF_seasons = ((LF.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')

#     A_seasons   = ((A.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')

#     C_seasons   = ((C.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')

#     N_seasons  = ((N.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
#     NE_seasons = ((NE.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
#     E_seasons  = ((E.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
#     SE_seasons = ((SE.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
#     S_seasons  = ((S.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
#     SW_seasons = ((SW.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
#     W_seasons  = ((W.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
#     NW_seasons = ((NW.resample(time = 'Q-FEB').sum() / len_seasons) *100).groupby('time.season').mean(dim = 'time')
    
#     CTs = {'LF':LF_seasons, 
#        'A': A_seasons, 
#       'C':C_seasons, 
#       'NE':NE_seasons, 'E':E_seasons, 'SE':SE_seasons, 'S':S_seasons, 'SW':SW_seasons, 'W':W_seasons, 'NW':NW_seasons, 'N':N_seasons}
#     seasons_labels=NW_seasons.season
#     lat_vals = NW_seasons.lat
#     lon_vals = NW_seasons.lon
#     seasonal = []
#     for i in CTs:
#         seasonal.append(CTs[i])
#     seasonal = np.array(seasonal)
#     seasonal_xr=xr.DataArray(seasonal,
#                   coords = {'CT': list(CTs.keys()),                        
#                             'season': seasons_labels,
#                            'lat': lat_vals,
#                            'lon': lon_vals},
#                   dims = ['CT', 'season', 'lat', 'lon'],
#                   attrs=dict(
#                         description="Seasonal relative frequencies period " + str(year_init)+'-'+str(year_end),
#                         units="%",
#                   ) )

#     seasonal_xr.name = 'rel_freq'
#     return(seasonal_xr)






#!/usr/bin/env python
# coding: utf-8

# In[ ]:
"""
@Author: Pedro Herrera-Lormendez
"""

import gc
import numpy as np
import pandas as pd
import xarray as xr
#Importing the directory where the neccesary functions are located
from jcclass import JC_functions #Functions that help compute the CTs

def JC_classification(filename):
    
    '''
    
    @authors: Herrera-Lormendez, Pedro & John, Amal
    TU Bergakademie Freiberg and CNRS/Météo-France
    
    This computes the gridded Jenkinson-Collison circulation types
    derived from the original Lamb Weather Types Classification
    
    Computation of circulation types employs Mean Sea Level Pressure data
    
    :param filename: str. name and directory of the MSLP file
    :return: DataArray of grided circulation types data as an xarray file
    '''
    if type(filename) == str:
        print('Reading filename: ', filename)
        #Reading the file
        DS = xr.open_dataset(filename)
        mslp = JC_functions.renaming_coords(DS)
        #If apply, extracting metadata from Global Models
        is_gcm = False
        try: 
            institution_id = DS.institution_id
            experiment_id  = DS.experiment_id
            source_id      = DS.source_id
            is_gcm         = True
        except:
            pass            
        DS.close()
        del DS
        gc.collect()
          
        mslp_dim_1 = mslp.dims[1]
        #Checking longitude coordinates to be - 180 to 180
        print('Checking if longitude coordinates are -180 to 180...')
        mslp = JC_functions.checking_lon_coords(mslp)
        print('Checking latitude coordinates values...')
        mslp = JC_functions.checking_lat_coords(mslp)

        #Computing factors based on grid size
        dif_lon = float((mslp.lon[1]-mslp.lon[0]))
        dif_lat = float((mslp.lat[1]-mslp.lat[0]))

        factor_lat = float(np.abs(1/dif_lat))
        factor_lon = float(np.abs(1/dif_lon))
        f_lat1 = int(10*factor_lat)
        f_lat2 = int(-10*factor_lat)
        
        f_lon1 = int(15*factor_lon)
        f_lon2 = int(-15*factor_lon)
        #Checking if the MSLP data covers the whole world
        answer_globe = JC_functions.is_world(mslp)
        if answer_globe == True:
            psl_area=mslp[:,f_lat1:f_lat2, :] 
        elif answer_globe == False:
            psl_area=mslp[:,f_lat1:f_lat2, f_lon1:f_lon2]
            
        lat = psl_area.lat
        lon = psl_area.lon
        time = psl_area.time.values
        del psl_area
        gc.collect()
        time_len = len(time)
        lon_list = list(lon)
        lat_list = list(lat)
        len_lon = len(lon_list)
        len_lat = len(lat_list)
        ds = []

        print('Calculating constants') 

        lat_central = lat
        phi = lat_central            
        constants = JC_functions.constants(phi, lon)
        sc = constants[0]
        zwa = constants[1]
        zwb = constants[2]
        zsc = constants[3]            


        print('Checking time formats')

        #Checking the time coordinate values 
        if type(time[0]) == np.datetime64:
            time_pd = pd.to_datetime(time)
            dates = [pd.to_datetime(str(time_pd[t].year) + '-' + str(time_pd[t].month) + '-' + str(time_pd[t].day), 
                                        format = '%Y%m%d',errors = 'ignore') for t in range(len(time))]
        else:
            dates = [pd.to_datetime(str(time[t].year) + '-' + str(time[t].month) + '-' + str(time[t].day), 
                                format = '%Y%m%d',errors = 'ignore') for t in range(len(time))]

        #Extracting the 16-gridded values of MSLP
        #extraction based on point 8 on original map
        print('Extracting the 16 MSLP gridpoints')
        if answer_globe == True:
            gridpoints = JC_functions.extracting_gridpoints_globe(mslp, lat, lon)
            p1  = gridpoints[0]
            p2  = gridpoints[1]
            p3  = gridpoints[2]
            p4  = gridpoints[3]
            p5  = gridpoints[4]
            p6  = gridpoints[5]
            p7  = gridpoints[6]
            p8  = gridpoints[7]
            p9  = gridpoints[8]
            p10 = gridpoints[9]
            p11 = gridpoints[10]
            p12 = gridpoints[11]
            p13 = gridpoints[12]
            p14 = gridpoints[13]
            p15 = gridpoints[14]
            p16 = gridpoints[15]
        elif answer_globe == False:
            gridpoints = JC_functions.extracting_gridpoints_area(mslp, lat, lon)
            p1  = gridpoints[0]
            p2  = gridpoints[1]
            p3  = gridpoints[2]
            p4  = gridpoints[3]
            p5  = gridpoints[4]
            p6  = gridpoints[5]
            p7  = gridpoints[6]
            p8  = gridpoints[7]
            p9  = gridpoints[8]
            p10 = gridpoints[9]
            p11 = gridpoints[10]
            p12 = gridpoints[11]
            p13 = gridpoints[12]
            p14 = gridpoints[13]
            p15 = gridpoints[14]
            p16 = gridpoints[15]
       
        print('Computing flow terms')
        #Computing equations of flows and vorticity                        
        flows = JC_functions.flows(p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, p16, sc, zwa, zsc, zwb, lat, lon, time, mslp)
        del (mslp, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, p16, sc, zwa, zsc, zwb)
        gc.collect()
        W  = flows[0] #Westerly flow
        S  = flows[1] #Southerly flow
        F  = flows[2] #Resultant flow
        ZW = flows[3] #Westerly shear vorticity
        ZS = flows[4] #Southerly shear vorticity
        Z  = flows[5] #Total shear vorticity

        del flows
        gc.collect()            

        print('Computing flow directions')
        #Computing the wind direction values          
        dd = np.arctan(W/S)
        deg = np.rad2deg(dd)
        deg=np.mod(180+np.rad2deg(np.arctan2(W, S)),360) 

        del (W, S, ZW, ZS)
        gc.collect()        
        #https://confluence.ecmwf.int/pages/viewpage.action?pageId=133262398 
        #Assigning the Wind Direction labels
        direction = JC_functions.direction_def_NH(deg)
        direction = xr.where(deg.lat < 0, JC_functions.direction_def_SH(deg), direction)            

        print('Determining the Circulation types')
        #Determination of Circulation Type (27 Original types)
        lwt = JC_functions.assign_lwt(F, Z, direction)

        del (F, Z, direction)
        gc.collect()
        
        #Storing the gridded Circulation Types in an xarray file
        print('Saving the data in an xarray format')
        if (mslp_dim_1 == 'latitude') or (mslp_dim_1 == 'lat'):
            output=xr.DataArray(data = lwt.values,
                coords = {'time': time,
                        'lat': lat_list, 
                        'lon': lon_list},
                        dims = ['time', 'lat', 'lon'])
            output.name = 'CT' #Assigning variable name
        elif mslp_dim_1 == 'number':
            output=xr.DataArray(data = lwt.values,
                coords = {'time': time,
                          'number':mslp.number,
                          'lat': lat_list, 
                          'lon': lon_list},
                        dims = ['time','number', 'lat', 'lon'])
            output.name = 'CT' #Assigning variable name
        if is_gcm == True:
            output.attrs = {
                'description':'Gridded Lamb circulation types derived from MSLP data based on the automated Jenkinson-Collison classification',
                'institution_id': institution_id,
                'source_id': source_id,
                'experiment_id': experiment_id}
        elif is_gcm== False:
            output.attrs = {
                'description':'Gridded Lamb circulation types derived from MSLP data based on the automated Jenkinson-Collison classification'}
        output = JC_functions.checking_lat_coords(output)
       
        print('The End!')


    else:
        raise TypeError("Incorrect filename directory. Only strings allowed")
    
    return output


#!/usr/bin/env python
# coding: utf-8

# In[ ]:
"""
@Author: Pedro Herrera-Lormendez
"""

import numpy as np
import pandas as pd
import xarray as xr

from jcclass import JC_functions #Functions that help compute the CTs
from jcclass import JC_classification
from jcclass import CTs_functions
from jcclass import CTs_plots


class jc:
    """
    This computes the gridded Lamb Weather Types
    using the Jenkinson-Collison automated classification
    
    ···
    Attributes
    ----------
    filename  : str
        path and name of the mean sea level pressure dataset as netcdf
    cts       : A 2D ['lat','lon'] xarray.DataArray 
        Derived 27 gridded circulation types
    mslp      : A 2D ['lat', 'lon'] xarray.Dataset
        Mean Sea Level Pressure dataset in Pa

    *args
    lat_south  : int, optional
        Northernmost latitude value (-90 to 90)
    lat_north  : int, optional
        Southernmost latitude value (-90 to 90)
    lon_west   : int, optinal
        Westernmost longitude value (-180 to 180)
    lon_east   : int, optional
        Easternmost latitude value (-180 to 180)
    *argsglobe        
    lat_central: int, optional. Default 30ºN
        Central latitude of plot (-90 to 90)
    lon_central: int, optional. Default = 0ºE  
        Central longitude of plot (-180 to 180)

    Methods
    -------
    jc(filename).classification():
        Computes the 27 gridded synoptic circulation types.

    eleven_cts(cts):
        Computes the reduced eleven CTs.

    plot_cts(cts, *args):
        Plots the reduced eleven CTs.

    plot_cts_mslp(cts, mslp, *args)
        Plots the reduced eleven CTs and MSLP contour lines.

    plot_cts_globe(cts, mslp, *argsglobe)
        Plots the reduced eleven CTs and MSLP contour lines 
        using a NearsidePerspective projection.

    """
    def __init__(self, filename):
        self.filename = filename
        
    def classification(self):
        ''' 
        Computes the 27 circulation types
        '''
        cts = JC_classification.JC_classification(self.filename)
        return (cts)
    @staticmethod    
    def eleven_cts(cts):
        '''
        Computes the reduced 11 circulation types
        '''
        cts_11 = CTs_functions.eleven_CTs(cts)
        return(cts_11)
    @staticmethod
    def plot_cts(cts, *args):
        '''
        Plots the circulation types over a region. 
        Default area globe
        '''
        fig = CTs_plots.plot_CT(cts, *args)
        return(fig)
    @staticmethod
    def plot_cts_mslp(cts, mslp, *args):
        '''
        Plots the circulations types and the 
        contour lines of MSLP over a region.
        Default area globe
        '''
        fig = CTs_plots.plot_CT_MSLP(cts, mslp, *args)
        return(fig)
    @staticmethod    
    def plot_cts_globe(cts, mslp, *argsglobe):
        '''
        Plots the circulation types and MSLP contour lines
        over the Globe using the NearsidePerspective projection
        '''
        fig = CTs_plots.plot_CT_MSLP_globe(cts,mslp, *argsglobe)
        return(fig)



#!/usr/bin/env python
# coding: utf-8

# In[1]:
"""
@Author: Pedro Herrera-Lormendez
"""

import numpy as np
import pandas as pd
import xarray as xr
#from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
import cftime
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import matplotlib.colors as colors
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
import cartopy.crs as ccrs
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
import matplotlib.ticker as mticker

from jcclass import JC_functions
from jcclass import CTs_functions

def plot_CT(CT, lat_south=-80, lat_north=80, lon_west=-180, lon_east=180):
    '''
    This function plots the 11 circulation types to a map
    
    :param CT: A 2D ['lat','lon'] DataArray file of the 27 circulation types
    *args
    :param lat_south: int -80 to 80
    :param lat_north: int -80 to 80
    :param lon_west : int -180 to 180
    :param lon_east : int -180 to 180
    '''
    lat_north = CT.lat.sel(lat = lat_north, method = 'nearest')
    lat_south = CT.lat.sel(lat = lat_south, method = 'nearest')
    lon_west = CT.lon.sel(lon = lon_west, method = 'nearest')
    lon_east = CT.lon.sel(lon = lon_east, method = 'nearest')
    CT = CT.sel(lat = slice(lat_south, lat_north), lon = slice(lon_west, lon_east))        
    dif_x = lon_east - (lon_west)
    dif_y = lat_north - (lat_south)
    size_x = 12
    size_y = 10
    if dif_x > dif_y:
        size_x = 12
        size_y = size_x * (dif_y / dif_x)
    elif dif_x < dif_y:
        size_y = 12
        size_x = size_y * (dif_y / dif_x)
    
    CT = xr.where((CT.lat < 10) & (CT.lat > -10), np.nan, CT)
    highs = xr.where( (CT>=1) & (CT<=8), 1, np.nan)
    lows = xr.where( (CT>=21) & (CT<=28), 1, np.nan)
    CT = CTs_functions.eleven_CTs(CT)    
    #Defining colours to plot CTs
    colores = ListedColormap(["#7C7C77", "#17344F", "#0255F4","#0F78ED", "#9E09EE", "#F6664C",
                         "#F24E64", "#D3C42D", "#2FC698", 
                         "#20E1D7", "#BD0000"])  #C, N, NW, W, SW, S, SE, E, NE, A, LF  
    bounds = np.arange(-1,11,1)
    norm = colors.BoundaryNorm(boundaries=bounds, ncolors=12)
    lons, lats = CT.lon, CT.lat
    #Defining size and map boundaries    
    plt.gcf().clear()   
    fig= plt.figure(figsize = (size_x,size_y))
    proj = ccrs.PlateCarree()
    ax = plt.axes(projection=proj)
    ax.set_extent([lon_west, lon_east, lat_south, lat_north], proj)
    im = plt.pcolor(lons, lats, CT,
                 transform=proj, norm = norm, cmap = colores)
    ima = plt.contourf(lons, lats, highs, colors = None, hatches = ['....'], alpha = 0)
    imc = plt.contourf(lons, lats, lows, colors = None, hatches = ['----'], alpha = 0)

    ax.coastlines('50m')
    ax.stock_img()

    gl = ax.gridlines(crs=proj)
    gl.xlabels_top = False
    gl.xlabels_bottom = True
    gl.ylabels_left = True
    gl.ylabels_right = False
    gl.xlines = False
    gl.ylines = False
    gl.ylocator = mticker.FixedLocator([-80, -60, -40, -20, 0, 20, 40, 60, 80])
    gl.xlocator = mticker.FixedLocator([-180,-120,-60, 0, 60, 120, 180])
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER
   
    #Legends
    ax = plt.gca()
    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1,
                     box.width, box.height * 0.9])
    legend_elements = [  Line2D([0], [0],marker ='o', color = '#7C7C77',markerfacecolor = None,linestyle='None', markersize = 17,label='LF'),
                         Line2D([0], [0], marker='o', color='#17344F', markerfacecolor=None, linestyle='None', markersize=17, label='A'),
                         Line2D([0], [0], marker='o', color='#0255F4', markerfacecolor=None, linestyle='None', markersize=17, label='NE'),
                         Line2D([0], [0], marker='o', color='#0F78ED', markerfacecolor=None, linestyle='None', markersize=17, label='E'),
                         Line2D([0], [0], marker='o', color='#9E09EE', markerfacecolor=None, linestyle='None', markersize=17, label='SE'),
                         Line2D([0], [0], marker='o', color='#F6664C', markerfacecolor=None, linestyle='None', markersize=17, label='S'),
                         Line2D([0], [0], marker='o', color='#F24E64', markerfacecolor=None, linestyle='None', markersize=17, label='SW'),
                         Line2D([0], [0], marker='o', color='#D3C42D', markerfacecolor=None, linestyle='None', markersize=17, label='W'),
                         Line2D([0], [0], marker='o', color='#2FC698', markerfacecolor=None, linestyle='None', markersize=17, label='NW'),
                         Line2D([0], [0], marker='o', color='#20E1D7', markerfacecolor=None, linestyle='None', markersize=17, label='N'),
                         Line2D([0], [0], marker='o', color='#BD0000', markerfacecolor=None, linestyle='None', markersize=17, label='C'),
                         Line2D([0], [0], marker=None, color ='#000000',markerfacecolor=None, linestyle='--', markersize=17, label='Partly Cyclonic'),
                         Line2D([0], [0], marker=None, color ='#000000',markerfacecolor=None, linestyle='dotted', markersize=17, label='Partly Anticylonic'),                       
                      ]

    ax.legend(handles=legend_elements,loc = 'center', bbox_to_anchor=(1.15, 0.50),frameon = False, prop={'size': 14}) #X,Y, LENGTH, WIDHT
    
    plt.title(str(CT.time.values), size = 16)
    plt.tight_layout()
    plt.ioff()
    return fig

def plot_CT_MSLP(CT, MSLP, lat_south=-80, lat_north=80, lon_west=-180, lon_east=180):
    '''
    This function plots the 11 circulation types to a map
    ¡! Works only with Basemap module
    
    :param CT: xarray file of the 27 circulation types
    :param MSLP: a 2D ['lat', 'lon'] DataArray of Mean Sea Level Pressure data in Pa
    *args
    :param lat_south: int -80 to 80
    :param lat_north: int -80 to 80
    :param lon_west : int -180 to 180
    :param lon_east : int -180 to 180
    '''
    MSLP = JC_functions.renaming_coords(MSLP)


    MSLP = JC_functions.checking_lon_coords(MSLP)
    MSLP = JC_functions.checking_lat_coords(MSLP)

    lat_north = CT.lat.sel(lat = lat_north, method = 'nearest')
    lat_south = CT.lat.sel(lat = lat_south, method = 'nearest')
    lon_west = CT.lon.sel(lon = lon_west, method = 'nearest')
    lon_east = CT.lon.sel(lon = lon_east, method = 'nearest')
    CT = CT.sel(lat = slice(lat_south, lat_north), lon = slice(lon_west, lon_east))
    dif_x = lon_east - (lon_west)
    dif_y = lat_north - (lat_south)
    size_x = 12
    size_y = 10
    if dif_x > dif_y:
        size_x = 12
        size_y = size_x * (dif_y / dif_x)
    elif dif_x < dif_y:
        size_y = 12
        size_x = size_y * (dif_y / dif_x)        
        
    MSLP = MSLP.sel(lat = slice(lat_south, lat_north), lon = slice(lon_west, lon_east))
    lat = MSLP.lat
    lon = MSLP.lon
    lon_list = list(lon)
    lat_list = list(lat)          
    CT = xr.where((CT.lat < 10) & (CT.lat > -10), np.nan, CT)
    highs = xr.where( (CT>=1) & (CT<=8), 1, np.nan)
    lows = xr.where( (CT>=21) & (CT<=28), 1, np.nan)
    CT = CTs_functions.eleven_CTs(CT)           

    #Defining colours to plot CTs
    colores = ListedColormap(["#7C7C77", "#17344F", "#0255F4","#0F78ED", "#9E09EE", "#F6664C",
                         "#F24E64", "#D3C42D", "#2FC698", 
                         "#20E1D7", "#BD0000"])  #C, N, NW, W, SW, S, SE, E, NE, A, LF  
    bounds = np.arange(-1,11,1)
    norm = colors.BoundaryNorm(boundaries=bounds, ncolors=12)
    lons, lats = np.meshgrid(CT.lon, CT.lat)
    #Defining size and map boundaries
    plt.gcf().clear()   
    fig= plt.figure(figsize = (size_x,size_y))
    proj = ccrs.PlateCarree()
    ax = plt.axes(projection=ccrs.PlateCarree())
    ax.set_extent([lon_west, lon_east, lat_south, lat_north], proj)
    ax.coastlines('50m')

    ax.stock_img()
    
    im = plt.pcolor(lons, lats, CT,
             transform=proj, norm = norm, cmap = colores, alpha = 0.75)
    from scipy.ndimage.filters import gaussian_filter
    data3 = gaussian_filter(MSLP, sigma=.8)    
    im2 = plt.contour(lons, lats, data3, 
                      np.arange(1012-40, 1012+44, 4),transform=proj, colors = '0.10', linewidths = 0.5)
    ima = plt.contourf(lons, lats, highs, colors = None, hatches = ['....'], alpha = 0)
    imc = plt.contourf(lons, lats, lows, colors = None, hatches = ['----'], alpha = 0)

    plt.clabel(im2, im2.levels, inline=True,fmt ='%.0f', fontsize=10)

    gl = ax.gridlines(crs=proj)
    gl.xlabels_top = False
    gl.xlabels_bottom = True
    gl.ylabels_left = True
    gl.ylabels_right = False
    gl.xlines = False
    gl.ylines = False
    gl.ylocator = mticker.FixedLocator([-80, -60, -40, -20, 0, 20, 40, 60, 80])
    gl.xlocator = mticker.FixedLocator([-180, -150,-120,-90,-60,-30, 0, 30, 60, 90, 120, 150, 180])
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER

    legend_elements = [  Line2D([0], [0],marker ='o', color = '#7C7C77',markerfacecolor = None,linestyle='None', markersize = 17,label='LF'),
                         Line2D([0], [0], marker='o', color='#17344F', markerfacecolor=None, linestyle='None', markersize=17, label='A'),
                         Line2D([0], [0], marker='o', color='#0255F4', markerfacecolor=None, linestyle='None', markersize=17, label='NE'),
                         Line2D([0], [0], marker='o', color='#0F78ED', markerfacecolor=None, linestyle='None', markersize=17, label='E'),
                         Line2D([0], [0], marker='o', color='#9E09EE', markerfacecolor=None, linestyle='None', markersize=17, label='SE'),
                         Line2D([0], [0], marker='o', color='#F6664C', markerfacecolor=None, linestyle='None', markersize=17, label='S'),
                         Line2D([0], [0], marker='o', color='#F24E64', markerfacecolor=None, linestyle='None', markersize=17, label='SW'),
                         Line2D([0], [0], marker='o', color='#D3C42D', markerfacecolor=None, linestyle='None', markersize=17, label='W'),
                         Line2D([0], [0], marker='o', color='#2FC698', markerfacecolor=None, linestyle='None', markersize=17, label='NW'),
                         Line2D([0], [0], marker='o', color='#20E1D7', markerfacecolor=None, linestyle='None', markersize=17, label='N'),
                         Line2D([0], [0], marker='o', color='#BD0000', markerfacecolor=None, linestyle='None', markersize=17, label='C'),
                Line2D([0], [0], marker=None, color ='#000000',markerfacecolor=None, linestyle='--', markersize=17, label='Partly Cyclonic'),
                         Line2D([0], [0], marker=None, color ='#000000',markerfacecolor=None, linestyle='dotted', markersize=17, label='Partly Anticylonic'),                       
                      ]

    ax.legend(handles=legend_elements,loc = 'center', bbox_to_anchor=(1.15, 0.50),frameon = False, prop={'size': 14}) #X,Y, LENGTH, WIDHT                         

    #plt.legend(handles=legend_elements, loc='right', bbox_to_anchor=(0.65, 0.25, 0.5, 0.5),frameon = False, prop={'size': 14}) #X,Y, LENGTH, WIDHT
    plt.title(str(CT.time.values), size = 16, y = 1.05)

    plt.tight_layout()
    plt.ioff()
    return fig
def plot_CT_MSLP_globe(CT, MSLP, lat_central = 30, lon_central=0):
    '''
    This function plots the 11 circulation types to a map
    ¡! Works only with Basemap module

    :param CT: xarray file of the 27 circulation types
    :param MSLP: a 2D ['lat', 'lon'] DataArray of Mean Sea Level Pressure data in Pa
    *args
    :param lat_central: int -80 to 80
    :param lon_central: int -80 to 80

    '''
    MSLP = JC_functions.renaming_coords(MSLP)

    MSLP = JC_functions.checking_lon_coords(MSLP)
    MSLP = JC_functions.checking_lat_coords(MSLP)

    lat_north = float(CT.lat[-1].values)
    lat_south = float(CT.lat[0].values)
    lon_west  = float(CT.lon[0].values)
    lon_east  = float(CT.lon[-1].values)        

    MSLP = MSLP.sel(lat = slice(lat_south, lat_north), lon = slice(lon_west, lon_east))
    lat = MSLP.lat
    lon = MSLP.lon
    lon_list = list(lon)
    lat_list = list(lat)          
    CT = xr.where((CT.lat < 10) & (CT.lat > -10), np.nan, CT)
    highs = xr.where( (CT>=1) & (CT<=8), 1, np.nan)
    lows = xr.where( (CT>=21) & (CT<=28), 1, np.nan)
    CT = CTs_functions.eleven_CTs(CT)

    #Defining colours to plot CTs
    colores = ListedColormap(["#7C7C77", "#17344F", "#0255F4","#0F78ED", "#9E09EE", "#F6664C",
                         "#F24E64", "#D3C42D", "#2FC698", 
                         "#20E1D7", "#BD0000"])  #C, N, NW, W, SW, S, SE, E, NE, A, LF  
    bounds = np.arange(-1,11,1)
    norm = colors.BoundaryNorm(boundaries=bounds, ncolors=12)
    lons, lats = np.meshgrid(CT.lon, CT.lat)
    #Defining size and map boundaries
    # set perspective angle
    fig= plt.figure(figsize = (12,10))
    proj = ccrs.Orthographic(central_longitude = lon_central, central_latitude = lat_central)
    ax = plt.axes(projection=proj)
    ax.coastlines('50m')
    ax.stock_img()
    im1 = plt.pcolor(lons, lats, CT,
                 transform=ccrs.PlateCarree(), norm = norm, cmap = colores, alpha = 0.75)
    from scipy.ndimage.filters import gaussian_filter
    data3 = gaussian_filter(MSLP, sigma=.8)   
    # data3 = np.round(data3,0)
    im2 = plt.contour(lons,lats,data3, 
                     np.arange(1012-40, 1012+44, 4), 
                      transform=ccrs.PlateCarree(), colors = '0.10', linewidths = 0.5)
    ima = plt.contourf(lons, lats, highs, transform=ccrs.PlateCarree(), colors = None, hatches = ['....'], alpha = 0)
    imc = plt.contourf(lons, lats, lows, transform=ccrs.PlateCarree(), colors = None, hatches = ['----'], alpha = 0)


    # plt.clabel(im2, im2.levels, manual = False, inline=True,fmt ='%.0f', fontsize=9)

    legend_elements = [  Line2D([0], [0],marker ='o', color = '#7C7C77',markerfacecolor = None,linestyle='None', markersize = 17,label='LF'),
                         Line2D([0], [0], marker='o', color='#17344F', markerfacecolor=None, linestyle='None', markersize=17, label='A'),
                         Line2D([0], [0], marker='o', color='#0255F4', markerfacecolor=None, linestyle='None', markersize=17, label='NE'),
                         Line2D([0], [0], marker='o', color='#0F78ED', markerfacecolor=None, linestyle='None', markersize=17, label='E'),
                         Line2D([0], [0], marker='o', color='#9E09EE', markerfacecolor=None, linestyle='None', markersize=17, label='SE'),
                         Line2D([0], [0], marker='o', color='#F6664C', markerfacecolor=None, linestyle='None', markersize=17, label='S'),
                         Line2D([0], [0], marker='o', color='#F24E64', markerfacecolor=None, linestyle='None', markersize=17, label='SW'),
                         Line2D([0], [0], marker='o', color='#D3C42D', markerfacecolor=None, linestyle='None', markersize=17, label='W'),
                         Line2D([0], [0], marker='o', color='#2FC698', markerfacecolor=None, linestyle='None', markersize=17, label='NW'),
                         Line2D([0], [0], marker='o', color='#20E1D7', markerfacecolor=None, linestyle='None', markersize=17, label='N'),
                         Line2D([0], [0], marker='o', color='#BD0000', markerfacecolor=None, linestyle='None', markersize=17, label='C'),
                         Line2D([0], [0], marker=None, color ='#000000',markerfacecolor=None, linestyle='--', markersize=17, label='Partly Cyclonic'),
                         Line2D([0], [0], marker=None, color ='#000000',markerfacecolor=None, linestyle='dotted', markersize=17, label='Partly Anticylonic'),                       
                      ]

    plt.legend(handles=legend_elements, loc='right', bbox_to_anchor=(0.90, 0.25, 0.5, 0.5),frameon = False, prop={'size': 14}) #X,Y, LENGTH, WIDHT

    plt.title(str(CT.time.values), size = 16, y = 1.05)

    plt.tight_layout()
    plt.ioff()
    return fig





from setuptools import find_packages, setup

'''
@Author: Pedro Herrera-Lormendez
'''

vfile = {}
exec(open('jcclass/version.py').read(), vfile)

with open('README.md', 'r') as fh:
	long_description = fh.read()

#with open('jc_class/requirements.txt', 'r') as fh:
#	install_requires = fh.read()


setup(
	name='jcclass',
	#version = '0.0.2',
	include_package_data = True,
	keywords='circulations, CTs, WTs, synoptic',
	author = 'Pedro Herrera-Lormendez',
	author_email='peth31@gmail.com',
	description = 'Jenkinson and Collison automated gridded classification',
	long_description = long_description,
    long_description_content_type='text/markdown',	
	url = 'https://github.com/PedroLormendez/jcclass',
	license = 'MIT',
	#packages =['jcclass'],
	packages=find_packages(),
	zip_safe = False,
	python_requires='>=3.7',
	setuptools_git_versioning ={
		'version_callback' : vfile['__version__'],
		'dev_template'  : '{tag}.posst{ccount}',
		'dirty_template': '{tag}.post{ccount}'
		},
	setup_requires = ['setuptools-git-versioning', 'numpy'],
	install_requires = [
						'numpy>=1.19.5',
						'xarray>=0.16.2',
						'matplotlib>=3.2.0',
						'pyproj',
						'cartopy>=0.17.0',
						'cftime',
						'netCDF4',
						#'h5netcdf',
	]
	)


#!/usr/bin/env python
# coding: utf-8

# In[ ]:
"""
@Author: Pedro Herrera-Lormendez
"""
#Importing neccesary modules
import gc
import numpy as np
import pandas as pd
import xarray as xr

def renaming_coords(DS):
    """
    This function reads and extracts the latitude and loongitude coordinate
    names as well as the name of the variable for MSLP. 
    Then read the variable and divides by 100 to convert to hPa
    If MSLP variable is different to 'msl' or 'psl', the user
    can input the name of the variable manually
    :param DS: Xarray dataset of MSLP.
    :return mslp: DataArray of MSLP data with lat,lon coordinates
    """
#Extracting name of longitude
    for lon_name in ('longitude', 'lon'):
        try:
            DS[lon_name]
            lon_name = lon_name
            break
        except KeyError:
            continue
    else:
        err_msg = (
                'Coodinate name for longitude different to "lon" or "longitude" '
            )
        raise KeyError(err_msg)
    #Extracting name of latitude    
    for lat_name in ('latitude', 'lat'):
        try:
            DS[lat_name]
            lat_name = lat_name
            break
        except KeyError:
            continue
    else:
        err_msg = (
                'Coodinate name for latitude different to "lat" or "latitude" '
            )
        raise KeyError(err_msg)

    #Renaming latitude, longitude coordinates to 'lat' and 'lon'
    if lat_name == 'latitude':
        DS = DS.rename({lat_name:'lat'})
    elif lat_name == 'lat':
        pass
    if lon_name == 'longitude':
        DS = DS.rename({lon_name:'lon'})
    elif lon_name == 'lon':
        pass
    #Reading variable
    for var_name in ('msl','psl'): #Usual names for MSLP variables
        try:
            mslp = DS[var_name]
            mslp /= 100
            break
        except KeyError:
            continue
    else: #If variable has a name different to 'msl' or 'psl'
        print('File contains a variable name different to "msl" or "psl".\n'
            'Please pvovide the variable name of the xr.Dataset containing '
            'sea level pressure:')
        var_name = input()
        mslp = DS[str(var_name)]
        mslp /= 100
    return(mslp)    

def checking_lat_coords(data):
    """
    This function checks and fixes the latitude coordinate
    values to go in incremental values (e.g., -80 to 80ºN)

    :param mslp: mean sea level pressure data in DataArray format
    """
    if data.lat[0] > data.lat[-1]:
        data = data.reindex(lat = list(reversed(data.lat)))
    else:
        pass
    return(data)

def checking_lon_coords(data):
    """
    This function checks and fixes the longitude coordinate
    values going from 180 to -180 º.
    
    :param mslp: mean sea level pressure data in DataArray format
    """
    lon_name = 'lon'
    if data[lon_name][-1] > 180:
        # Adjust lon values to make sure they are within (-180, 180)
        data['_longitude_adjusted'] = xr.where(
            data[lon_name] > 180,
            data[lon_name] - 360,
            data[lon_name])

        # reassign the new coords to as the main lon coords
        # and sort DataArray using new coordinate values
        data = (
            data
            .swap_dims({lon_name: '_longitude_adjusted'})
            .sel(**{'_longitude_adjusted': sorted(data._longitude_adjusted)})
            .drop(lon_name))

        data = data.rename({'_longitude_adjusted': lon_name})
    else:
        pass
    return(data)
def is_world(data):
    dif_lon = np.abs(data.lon[0] - data.lon[1])
    dif_lat = np.abs(data.lat[0] - data.lat[1])
    condition_east  = data.lon[-1]>= (180 - dif_lon) #East
    condition_west  = data.lon[0]<= (-180 + dif_lon) #West
    es_mundo = False
    if condition_west == condition_east:
        es_mundo = True
    else:
        pass
    return(es_mundo)


def constants(phi, lon):
    """
    Computing values of constants dependant on latitude and longitude
    They represent the constants referred to the relative differences
    between the grid-point spacing in the E-W and N-S direction
    
    :param phi: values of central latitude gridpoints
    :param lon: longitude values
    """
    SC = 1/np.cos(np.deg2rad(phi))
    SC.name="longitue"
    sc=xr.concat([SC]*len(lon),'logitude').T

    ZWA = np.sin(np.deg2rad(phi)) / np.sin (np.deg2rad(phi - 5))
    ZWA.name="longitue"
    zwa=xr.concat([ZWA]*len(lon),'logitude').T

    ZWB = np.sin(np.deg2rad(phi)) / np.sin (np.deg2rad(phi + 5))
    ZWB.name="longitue"
    zwb=xr.concat([ZWB]*len(lon),'logitude').T

    ZSC = (1/(2*(np.cos(np.deg2rad(phi))**2)))
    ZSC.name="longitue"
    zsc=xr.concat([ZSC]*len(lon),'logitude').T
    return (sc, zwa, zwb, zsc)

def direction_def_NH(deg_used):
    '''
    This function assigns the wind direction labels of the circulation types for 
    the Northern Hemisphere
    
    :param deg_used: xarray of wind direction values in degrees

    '''
    direction = xr.where( (deg_used>247) & (deg_used<=292), 'W', np.nan)
    direction = xr.where( (deg_used>292) & (deg_used<=337), 'NW', direction)
    direction = xr.where( (deg_used>337), 'N', direction)
    direction = xr.where( (deg_used>=0) & (deg_used<=22), 'N', direction)
    direction = xr.where( (deg_used>22) & (deg_used<=67), 'NE', direction)
    direction = xr.where( (deg_used>67) & (deg_used<=112), 'E', direction)
    direction = xr.where( (deg_used>112) & (deg_used<=157), 'SE', direction)
    direction = xr.where( (deg_used>157) & (deg_used<=202), 'S', direction)
    direction = xr.where( (deg_used>202) & (deg_used<=247), 'SW', direction)
    return(direction)

def direction_def_SH(deg_used):
    '''
    This function assigns the wind direction labels of the circulation types for 
    the Southern Hemisphere
    
    :param deg_used: xarray of wind direction values in degrees
    '''    
    direction = xr.where( (deg_used>247) & (deg_used<=292), 'E', np.nan)
    direction = xr.where( (deg_used>292) & (deg_used<=337), 'SE', direction)
    direction = xr.where( (deg_used>337), 'S', direction)
    direction = xr.where( (deg_used>=0) & (deg_used<=22), 'S', direction)
    direction = xr.where( (deg_used>22) & (deg_used<=67), 'SW', direction)
    direction = xr.where( (deg_used>67) & (deg_used<=112), 'W', direction)
    direction = xr.where( (deg_used>112) & (deg_used<=157), 'NW', direction)
    direction = xr.where( (deg_used>157) & (deg_used<=202), 'N', direction)
    direction = xr.where( (deg_used>202) & (deg_used<=247), 'NE', direction)
    return(direction)

def assign_lwt(F_i, Z_i, direction_i):
    '''
    This function assigns the corresponding circulation types' coding
    
    :param         F_i: xarray of Total Flow term (F)
    :param         Z_i: xarry of Total Shear Vorticity term (Z)
    :param direction_i: xarray of Flow Direction
    '''

    ### Hybrid Anticyclonic flows ###
    lwt = xr.where( (Z_i<0) & (direction_i=='NE'), 1, np.nan)    
    lwt = xr.where( (Z_i<0) & (direction_i=='E'),  2, lwt)
    lwt = xr.where( (Z_i<0) & (direction_i=='SE'), 3, lwt)
    lwt = xr.where( (Z_i<0) & (direction_i=='S'),  4, lwt)
    lwt = xr.where( (Z_i<0) & (direction_i=='SW'), 5, lwt)
    lwt = xr.where( (Z_i<0) & (direction_i=='W'),  6, lwt)
    lwt = xr.where( (Z_i<0) & (direction_i=='NW'), 7, lwt)
    lwt = xr.where( (Z_i<0) & (direction_i=='N'),  8, lwt)
    ### Hybrid Cyclonic flows ###
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='NE'), 11, lwt)    
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='E'),  12, lwt)
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='SE'), 13, lwt)
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='S'),  14, lwt)
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='SW'), 15, lwt)
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='W'),  16, lwt)
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='NW'), 17, lwt)
    lwt = xr.where( (np.absolute(Z_i)<F_i) & (direction_i=='N'),  18, lwt)
    ### Purely Cyclonic ###
    lwt = xr.where( ( (np.absolute(Z_i)) > (2*F_i) ) & (Z_i>0), 20, lwt)
    ### Purely Anticyclonic ###
    lwt = xr.where( ( (np.absolute(Z_i)) > (2*F_i) ) & (Z_i<0),  0, lwt)
    ### Directional flows ###
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='NE'), 21, lwt)    
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='E'),  22, lwt)
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='SE'), 23, lwt)
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='S'),  24, lwt)
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='SW'), 25, lwt)
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='W'),  26, lwt)
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='NW'), 27, lwt)
    lwt = xr.where( (np.absolute(Z_i)>F_i) & (np.absolute(Z_i) < 2*F_i) & (Z_i>0) & (direction_i=='N'),  28, lwt)
    ### Low Flow / Unclassified / Weak Flow ###
    lwt = xr.where( (F_i<6) & (np.absolute(Z_i) < 6), -1, lwt)

    return lwt
    
def extracting_gridpoints_area(mslp, lat, lon):
    """
    This function extracts the 16 moving gridded points over a defined area (not globe)
    given a Reanalysis or Global Climate Model Dataset. This gridpoints are
    neccesary for the computation of the terms
    """
    #Gridpoint 1
    lat1 = lat+10
    lon1 = lon-5
    lat1 = lat.sel(lat = lat1, method = 'nearest')        
    lon1 = lon.sel(lon = lon1, method = 'nearest')     
    p1 = np.array(mslp.sel(lat = lat1, lon = lon1))
    del(lat1, lon1)
    gc.collect()
    #Gridpoint 2
    lat2 = lat + 10
    lon2 = lon + 5
    lat2 = lat.sel(lat = lat2, method = 'nearest')        
    lon2 = lon.sel(lon = lon2, method = 'nearest')    
    p2 = np.array(mslp.sel(lat = lat2, lon = lon2))
    del (lat2, lon2)
    gc.collect()
    #Gridpoint 3
    lat3 = lat + 5
    lon3 = lon - 15
    lat3 = lat.sel(lat = lat3, method = 'nearest')        
    lon3 = lon.sel(lon = lon3, method = 'nearest')    
    p3 = np.array(mslp.sel(lat = lat3, lon = lon3))
    del (lat3, lon3)
    gc.collect()
    #Gridpoint 4
    lat4 = lat + 5
    lon4 = lon  -5
    lat4 = lat.sel(lat = lat4, method = 'nearest')        
    lon4 = lon.sel(lon = lon4, method = 'nearest')    
    p4 = np.array(mslp.sel(lat = lat4, lon = lon4))
    del (lat4, lon4)
    gc.collect()
    #Gridpoint 5
    lat5 = lat + 5
    lon5 = lon + 5
    lat5 = lat.sel(lat = lat5, method = 'nearest')        
    lon5 = lon.sel(lon = lon5, method = 'nearest')    
    p5 = np.array(mslp.sel(lat = lat5, lon = lon5))
    del (lat5, lon5)
    gc.collect()
    #Gridpoint 6
    lat6 = lat + 5
    lon6 = lon + 15
    lat6 = lat.sel(lat = lat6, method = 'nearest')        
    lon6 = lon.sel(lon = lon6, method = 'nearest')    
    p6 = np.array(mslp.sel(lat = lat6, lon = lon6))
    del (lat6, lon6)
    gc.collect()
    #Gridpoint 7
    lat7 = lat
    lon7 = lon - 15
    lat7 = lat.sel(lat = lat7, method = 'nearest')        
    lon7 = lon.sel(lon = lon7, method = 'nearest')    
    p7 = np.array(mslp.sel(lat = lat7, lon = lon7))
    del (lat7, lon7)
    gc.collect()
    #Gridpoint 8
    lat8 = lat
    lon8 = lon -5
    lat8 = lat.sel(lat = lat8, method = 'nearest')        
    lon8 = lon.sel(lon = lon8, method = 'nearest')    
    p8 = np.array(mslp.sel(lat = lat8, lon = lon8))
    del (lat8, lon8)
    gc.collect()
    #Gridpoint 9
    lat9 = lat
    lon9 = lon + 5
    lat9 = lat.sel(lat = lat9, method = 'nearest')        
    lon9 = lon.sel(lon = lon9, method = 'nearest')    
    p9 = np.array(mslp.sel(lat = lat9, lon = lon9))
    del (lat9, lon9)
    gc.collect()
    #Gridpoint 10
    lat10 = lat
    lon10 = lon + 15
    lat10 = lat.sel(lat = lat10, method = 'nearest')        
    lon10 = lon.sel(lon = lon10, method = 'nearest')    
    p10 = np.array(mslp.sel(lat = lat10, lon = lon10))
    del (lat10, lon10)
    gc.collect()
    #Gridpoint 11
    lat11 = lat - 5
    lon11 = lon - 15
    lat11 = lat.sel(lat = lat11, method = 'nearest')        
    lon11 = lon.sel(lon = lon11, method = 'nearest')    
    p11 = np.array(mslp.sel(lat = lat11, lon = lon11))
    del (lat11, lon11)
    gc.collect()
    #Gridpoint 12
    lat12 = lat - 5
    lon12 = lon - 5
    lat12 = lat.sel(lat = lat12, method = 'nearest')        
    lon12 = lon.sel(lon = lon12, method = 'nearest')    
    p12 = np.array(mslp.sel(lat = lat12, lon = lon12))
    del (lat12, lon12)
    gc.collect()
    #Gridpoint 13
    lat13 = lat - 5
    lon13 = lon + 5
    lat13 = lat.sel(lat = lat13, method = 'nearest')        
    lon13 = lon.sel(lon = lon13, method = 'nearest')    
    p13 = np.array(mslp.sel(lat = lat13, lon = lon13))
    del (lat13, lon13)
    gc.collect()
    #Gridpoint 14
    lat14 = lat - 5
    lon14 = lon + 15
    lat14 = lat.sel(lat = lat14, method = 'nearest')        
    lon14 = lon.sel(lon = lon14, method = 'nearest')    
    p14 = np.array(mslp.sel(lat = lat14, lon = lon14))
    del (lat14, lon14)
    gc.collect()
    #Gridpoint 15
    lat15 = lat - 10
    lon15 = lon - 5
    lat15 = lat.sel(lat = lat15, method = 'nearest')        
    lon15 = lon.sel(lon = lon15, method = 'nearest')    
    p15 = np.array(mslp.sel(lat = lat15, lon = lon15))
    del (lat15, lon15)
    gc.collect()
    #Gridpoint 16
    lat16 = lat - 10
    lon16 = lon + 5
    lat16 = lat.sel(lat = lat16, method = 'nearest')        
    lon16 = lon.sel(lon = lon16, method = 'nearest')    
    p16 = np.array(mslp.sel(lat = lat16, lon = lon16))
    del (lat16, lon16)
    gc.collect()
    return (p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, p16)
    
def extracting_gridpoints_globe(mslp, lat, lon):
    """
    This function extracts the 16 moving gridded points over the whole globe
    given a Reanalysis or Global Climate Model dataset. These gridpoints are
    neccesary for the computation fo the terms
    """
    #Gridpoint 1
    lat1 = lat + 10
    lon1 = xr.where(lon<-175, 360+lon-5, lon-5)
    lat1 = lat.sel(lat = lat1, method = 'nearest')        
    lon1 = lon.sel(lon = lon1, method = 'nearest')    
    p1 = np.array(mslp.sel(lat = lat1, lon = lon1))
    del (lat1, lon1)
    gc.collect()

    #Gridpoint 2
    lat2 = lat + 10
    lon2 = xr.where(lon>175, lon+5-360, lon+5)
    lon2 = xr.where(lon2 == 180, -180, lon2)
    lat2 = lat.sel(lat = lat2, method = 'nearest')        
    lon2 = lon.sel(lon = lon2, method = 'nearest')    
    p2 = np.array(mslp.sel(lat = lat2, lon = lon2))
    del (lat2, lon2)
    gc.collect()

    #Gridpoint 3
    lat3 = lat + 5
    lon3 = xr.where(lon<-165, 360+lon-15, lon-15)
    lat3 = lat.sel(lat = lat3, method = 'nearest')        
    lon3 = lon.sel(lon = lon3, method = 'nearest')    
    p3 = np.array(mslp.sel(lat = lat3, lon = lon3))
    del (lat3, lon3)
    gc.collect()    

    #Gridpoint 4
    lat4 = lat + 5
    lon4 = xr.where(lon<-175, 360+lon-5, lon-5)
    lat4 = lat.sel(lat = lat4, method = 'nearest')        
    lon4 = lon.sel(lon = lon4, method = 'nearest')    
    p4 = np.array(mslp.sel(lat = lat4, lon = lon4))
    del (lat4, lon4)
    gc.collect()    

    #Gridpoint 5
    lat5 = lat + 5
    lon5 = xr.where(lon>175, lon+5-360, lon+5)
    lon5 = xr.where(lon5 == 180, -180, lon5)
    lat5 = lat.sel(lat = lat5, method = 'nearest')        
    lon5 = lon.sel(lon = lon5, method = 'nearest')    
    p5 = np.array(mslp.sel(lat = lat5, lon = lon5))
    del (lat5, lon5)
    gc.collect()    

    #Gridpoint 6
    lat6 = lat + 5
    lon6 = xr.where(lon>165, lon+15-360,lon+15)
    lon6 = xr.where(lon6 == 180, -180, lon6)
    lat6 = lat.sel(lat = lat6, method = 'nearest')        
    lon6 = lon.sel(lon = lon6, method = 'nearest')    
    p6 = np.array(mslp.sel(lat = lat6, lon = lon6))
    del (lat6, lon6)
    gc.collect()    

    #Gridpoint 7
    lat7 = lat
    lon7 = xr.where(lon<-165, 360+lon-15, lon-15)
    lat7 = lat.sel(lat = lat7, method = 'nearest')        
    lon7 = lon.sel(lon = lon7, method = 'nearest')    
    p7 = np.array(mslp.sel(lat = lat7, lon = lon7))
    del (lat7, lon7)
    gc.collect()    

    #Gridpoint 8
    lat8 = lat
    lon8 = xr.where(lon<-175, 360+lon-5, lon-5)
    lat8 = lat.sel(lat = lat8, method = 'nearest')        
    lon8 = lon.sel(lon = lon8, method = 'nearest')    
    p8 = np.array(mslp.sel(lat = lat8, lon = lon8))
    del (lat8, lon8)
    gc.collect()    

    #Gridpoint 9
    lat9 = lat
    lon9 = xr.where(lon>175, lon+5-360, lon+5)
    lon9 = xr.where(lon9 == 180, -180, lon9)
    lat9 = lat.sel(lat = lat9, method = 'nearest')        
    lon9 = lon.sel(lon = lon9, method = 'nearest')    
    p9 = np.array(mslp.sel(lat = lat9, lon = lon9))
    del (lat9, lon9)
    gc.collect()    

    #Gridpoint 10
    lat10 = lat
    lon10 = xr.where(lon>165, lon+15-360, lon+15)
    lon10 = xr.where(lon10 == 180, -180, lon10)
    lat10 = lat.sel(lat = lat10, method = 'nearest')        
    lon10 = lon.sel(lon = lon10, method = 'nearest')    
    p10 = np.array(mslp.sel(lat = lat10, lon = lon10))
    del (lat10, lon10)
    gc.collect()    

    #Gridpoint 11
    lat11 = lat - 5
    lon11 = xr.where(lon<-165, 360+lon-15, lon-15)
    lat11 = lat.sel(lat = lat11, method = 'nearest')        
    lon11 = lon.sel(lon = lon11, method = 'nearest')    
    p11 = np.array(mslp.sel(lat = lat11, lon = lon11))
    del (lat11, lon11)
    gc.collect()        

    #Gridpoint 12
    lat12 = lat - 5
    lon12 = xr.where(lon<-175, 360+lon-5, lon-5)
    lat12 = lat.sel(lat = lat12, method = 'nearest')        
    lon12 = lon.sel(lon = lon12, method = 'nearest')    
    p12 = np.array(mslp.sel(lat = lat12, lon = lon12))
    del (lat12, lon12)
    gc.collect()        

    #Gridpoint 13
    lat13 = lat - 5 
    lon13 = xr.where(lon>175, lon+5-360, lon+5)
    lon13 = xr.where(lon13 == 180, -180, lon13)
    lat13 = lat.sel(lat = lat13, method = 'nearest')        
    lon13 = lon.sel(lon = lon13, method = 'nearest')    
    p13 = np.array(mslp.sel(lat = lat13, lon = lon13))
    del (lat13, lon13)
    gc.collect()        

    #Gridpoint 14
    lat14 = lat - 5
    lon14 = xr.where(lon>165, lon+15-360, lon+15)
    lon14 = xr.where(lon14 == 180, -180, lon14)
    lat14 = lat.sel(lat = lat14, method = 'nearest')        
    lon14 = lon.sel(lon = lon14, method = 'nearest')    
    p14 = np.array(mslp.sel(lat = lat14, lon = lon14))
    del (lat14, lon14)
    gc.collect()        

    #Gridpoint 15
    lat15 = lat - 10
    lon15 = xr.where(lon<-175, 360+lon-5, lon-5)
    lat15 = lat.sel(lat = lat15, method = 'nearest')        
    lon15 = lon.sel(lon = lon15, method = 'nearest')    
    p15 = np.array(mslp.sel(lat = lat15, lon = lon15))
    del (lat15, lon15)
    gc.collect()        

    #Gridpoint 16
    lat16 = lat - 10
    lon16 = xr.where(lon>175, lon+5-360, lon+5)
    lon16 = xr.where(lon16 == 180, -180, lon16)
    lat16 = lat.sel(lat = lat16, method = 'nearest')        
    lon16 = lon.sel(lon = lon16, method = 'nearest')    
    p16 = np.array(mslp.sel(lat = lat16, lon = lon16))
    del (lat16, lon16)
    gc.collect()        
    
    return (p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, p16)

def flows(p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, p16, sc, zwa, zsc, zwb, lat, lon, time, mslp):
    """
    This function computes the indices associated with the direction and vorticity
    of geostrophic flow given a reanalyses or GCM dataset
    More info: Jones, P. D., Hulme, M., & Briffa, K. R. (1993). 
    A comparison of Lamb circulation types with an objective classification scheme.  
    International Journal of Climatology(6), 655–663. https://doi.org/10.1002/joc.3370130606
    
    The function employs as parameters: 
    - the 16 gridded points of MSLP (p1 to p16)
    - The latitude dependant constants (sc, zwa, zsc  and zwb)
    - latitude, longitude, time and MSLP data
    
    This function also makes it possible to work with a dataset with an extra
    dimension named "number" which reffers to the number of a set of an ensemble. 
    This can be used for subseasonal forecasts coming from 
    the Climate Data Store: https://cds.climate.copernicus.eu/cdsapp#!/home
    """
    #Westerly Flow
    W = ((0.5)*( p12 + p13 )) - ((0.5)*( p4 + p5 ))
    if mslp.dims[1] == 'lat':
        W = xr.DataArray(W, 
                          coords = {'time': time, 'lat':lat, 'lon':lon}, 
                          dims = ['time', 'lat', 'lon'])
    elif mslp.dims[1] == 'number':
        W = xr.DataArray(W, 
                          coords = {'time': time, 'number':mslp.number ,'lat':lat, 'lon':lon}, 
                          dims = ['time','number' ,'lat', 'lon'])
        
    #Southerly Flow
    S = np.array(sc)*(((0.25)*(p5 + (2 * p9) + p13)) - ((0.25)*(p4 + (2 * p8) + p12)))
    if mslp.dims[1] == 'lat':
        S = xr.DataArray(S, 
                          coords = {'time': time, 'lat':lat, 'lon':lon}, 
                          dims = ['time', 'lat', 'lon'])
    elif mslp.dims[1] == 'number':
        S = xr.DataArray(S, 
                          coords = {'time': time, 'number':mslp.number ,'lat':lat, 'lon':lon}, 
                          dims = ['time','number', 'lat', 'lon'])        
    #Resultant Flow 
    F = np.sqrt(S**2 + W**2)
    #Westerly Shear Vorticity
    ZW = (np.array(zwa)*( (0.5)*(p15 + p16) - (0.5)*(p8 + p9))) - (np.array(zwb)*((0.5)*(p8 + p9) - (0.5)*(p1 + p2)))
    if mslp.dims[1] == 'lat':
        ZW = xr.DataArray(ZW, 
                          coords = {'time': time, 'lat':lat, 'lon':lon}, 
                          dims = ['time', 'lat', 'lon'])
    elif mslp.dims[1] == 'number':
        ZW = xr.DataArray(ZW, 
                              coords = {'time': time, 'number':mslp.number ,'lat':lat, 'lon':lon}, 
                              dims = ['time', 'number', 'lat', 'lon'])        
    #Southerly Shear Vorticity
    ZS = np.array(zsc) * ( ((0.25)*(p6 + (2 * p10) + p14)) - ((0.25)*(p5 + (2 * p9) + p13)) -((0.25)*(p4 + (2 * p8) + p12)) +((0.25)*(p3 + (2 * p7) + p11)) )
    #Total Shear Vorticity
    if mslp.dims[1] == 'lat':    
        ZS = xr.DataArray(ZS, 
                          coords = {'time': time, 'lat':lat, 'lon':lon}, 
                          dims = ['time', 'lat', 'lon'])
    elif mslp.dims[1] == 'number':
        ZS = xr.DataArray(ZS, 
                          coords = {'time': time,'number':mslp.number ,'lat':lat, 'lon':lon}, 
                          dims = ['time', 'number', 'lat', 'lon'])
        
    Z = ZW + ZS
    return(W, S, F, ZW, ZS, Z)
